@@ -1,37 +1,28 @@
+from __future__ import annotations
+
 import copy
 import json
 import logging
 import math
 import os
 import string
-import traceback
-from pathlib import Path
-from typing import (
-    Iterable,
-    Literal,
-    Optional,
-)
+from typing import Iterable, Literal, Optional
 
 import numpy as np
 import torch
 from numpy import ndarray
 from scipy.cluster import hierarchy
 from sentence_transformers import SentenceTransformer
-from sentence_transformers.model_card import (
-    SentenceTransformerModelCardData,
-    generate_model_card,
-)
 from sentence_transformers.models import Dense as DenseSentenceTransformer
 from sentence_transformers.models import Transformer
 from sentence_transformers.quantization import quantize_embeddings
-from sentence_transformers.similarity_functions import SimilarityFunction
-from sentence_transformers.util import (
-    batch_to_device,
-    load_file_path,
-)
+from sentence_transformers.util import batch_to_device, load_file_path
 from torch import nn
 from tqdm.autonotebook import trange
+from transformers.utils import cached_file
 
+from ..hf_hub.model_card import PylateModelCardData
+from ..scores import SimilarityFunction
 from ..utils import _start_multi_process_pool
 from .Dense import Dense
 
@@ -105,7 +96,8 @@ class ColBERT(SentenceTransformer):
         Whether to attend to the expansion tokens in the attention layers model. If False, the original tokens will
         not only attend to the expansion tokens, only the expansion tokens will attend to the original tokens. Default
         is False (as in the original ColBERT codebase).
-
+    skiplist_words
+        A list of words to skip from the documents scoring (note that these tokens are used for encoding and are only skipped during the scoring). Default is the list of string.punctuation.
     model_kwargs : dict, optional
         Additional model configuration parameters to be passed to the Huggingface Transformers model. Particularly
         useful options are:
@@ -212,28 +204,28 @@ class ColBERT(SentenceTransformer):
         truncate_dim: int | None = None,
         embedding_size: int | None = None,
         bias: bool = False,
-        query_prefix: str | None = "[Q] ",
-        document_prefix: str | None = "[D] ",
+        query_prefix: str | None = None,
+        document_prefix: str | None = None,
         add_special_tokens: bool = True,
         truncation: bool = True,
         query_length: int | None = None,
         document_length: int | None = None,
-        attend_to_expansion_tokens: bool = False,
+        attend_to_expansion_tokens: bool | None = None,
         skiplist_words: list[str] | None = None,
         model_kwargs: dict | None = None,
         tokenizer_kwargs: dict | None = None,
         config_kwargs: dict | None = None,
-        model_card_data: Optional[SentenceTransformerModelCardData] = None,
+        model_card_data: PylateModelCardData | None = None,
     ) -> None:
-        model_kwargs = {} if model_kwargs is None else model_kwargs
-        model_kwargs["add_pooling_layer"] = False
-
         self.query_prefix = query_prefix
         self.document_prefix = document_prefix
         self.query_length = query_length
         self.document_length = document_length
         self.attend_to_expansion_tokens = attend_to_expansion_tokens
         self.skiplist_words = skiplist_words
+        model_card_data = model_card_data or PylateModelCardData()
+        if similarity_fn_name is None:
+            similarity_fn_name = "MaxSim"
 
         super(ColBERT, self).__init__(
             model_name_or_path=model_name_or_path,
@@ -254,20 +246,74 @@ class ColBERT(SentenceTransformer):
             config_kwargs=config_kwargs,
             model_card_data=model_card_data,
         )
-
         hidden_size = self[0].get_word_embedding_dimension()
 
         # Add a linear projection layer to the model in order to project the embeddings to the desired size.
         if len(self) < 2:
-            # Add a linear projection layer to the model in order to project the embeddings to the desired size
-            embedding_size = embedding_size or 128
+            # If the model is a stanford-nlp ColBERT, load the weights of the dense layer
+            if (
+                self[0].auto_model.config.architectures is not None
+                and self[0].auto_model.config.architectures[0] == "HF_ColBERT"
+            ):
+                self.append(
+                    Dense.from_stanford_weights(
+                        model_name_or_path,
+                        cache_folder,
+                        revision,
+                        local_files_only,
+                        token,
+                        use_auth_token,
+                    )
+                )
+                logger.info("Loaded the weights from Stanford NLP model.")
+                try:
+                    metadata = cached_file(
+                        model_name_or_path,
+                        filename="artifact.metadata",
+                        cache_dir=cache_folder,
+                        revision=revision,
+                        local_files_only=local_files_only,
+                        token=token,
+                        use_auth_token=use_auth_token,
+                    )
+                    with open(metadata, "r") as f:
+                        metadata = json.load(f)
+                        # If the user do not override the values, read from config file
+                        if self.query_prefix is None:
+                            self.query_prefix = metadata["query_token_id"]
+                        if self.document_prefix is None:
+                            self.document_prefix = metadata["doc_token_id"]
+                        if self.query_length is None:
+                            self.query_length = metadata["query_maxlen"]
+                        if self.document_length is None:
+                            self.document_length = metadata["doc_maxlen"]
+                        if self.attend_to_expansion_tokens is None:
+                            self.attend_to_expansion_tokens = metadata[
+                                "attend_to_mask_tokens"
+                            ]
+                    logger.info("Loaded the configuration from Stanford NLP model.")
+                except EnvironmentError:
+                    if self.query_prefix is None:
+                        self.query_prefix = "[unused0]"
+                    if self.document_prefix is None:
+                        self.document_prefix = "[unused1]"
+                    # We do not set the query/doc length as they'll be set to the default values afterwards. We do it for prefixes as the default from stanford is different from ours
+                    logger.warning(
+                        "Could not load the configuration file from Stanford NLP model, using default values."
+                    )
+            else:
+                # Add a linear projection layer to the model in order to project the embeddings to the desired size
+                embedding_size = embedding_size or 128
 
-            logger.warning(
-                f"The checkpoint does not contain a linear projection layer. Adding one with output dimensions ({hidden_size}, {embedding_size})."
-            )
-            self.append(
-                Dense(in_features=hidden_size, out_features=embedding_size, bias=bias)
-            )
+                logger.info(
+                    f"The checkpoint does not contain a linear projection layer. Adding one with output dimensions ({hidden_size}, {embedding_size})."
+                )
+                logger.info("Created a PyLate model from base encoder.")
+                self.append(
+                    Dense(
+                        in_features=hidden_size, out_features=embedding_size, bias=bias
+                    )
+                )
 
         elif (
             embedding_size is not None
@@ -286,13 +332,36 @@ class ColBERT(SentenceTransformer):
             )
             self[1] = Dense.from_sentence_transformers(dense=self[1])
         else:
-            logger.warning("Pylate model loaded successfully.")
+            logger.info("PyLate model loaded successfully.")
+
+        # Ensure all tensors in the model are of the same dtype as the first tensor
+        try:
+            dtype = next(self.parameters()).dtype
+            self.to(dtype)
+        except StopIteration:
+            pass
 
         self.to(device)
         self.is_hpu_graph_enabled = False
+        # Override the configuration values with the provided arguments, if any. If not set and values have not been read from configs, set to default values.
+        self.query_prefix = (
+            query_prefix if query_prefix is not None else self.query_prefix or "[Q] "
+        )
+        self.document_prefix = (
+            document_prefix
+            if document_prefix is not None
+            else self.document_prefix or "[D] "
+        )
 
-        self.tokenizer.add_tokens([self.query_prefix, self.document_prefix])
-        self._first_module().auto_model.resize_token_embeddings(len(self.tokenizer))
+        # Try adding the prefixes to the tokenizer. We call resize_token_embeddings twice to ensure the tokens are added only if resize_token_embeddings works. There should be a better way to do this.
+        try:
+            self._first_module().auto_model.resize_token_embeddings(len(self.tokenizer))
+            self.tokenizer.add_tokens([self.query_prefix, self.document_prefix])
+            self._first_module().auto_model.resize_token_embeddings(len(self.tokenizer))
+        except NotImplementedError:
+            logger.warning(
+                "The tokenizer does not support resizing the token embeddings, the prefixes token have not been added to vocabulary."
+            )
 
         self.document_prefix_id = self.tokenizer.convert_tokens_to_ids(
             self.document_prefix
@@ -304,7 +373,6 @@ class ColBERT(SentenceTransformer):
         # Set the padding token ID to be the same as the mask token ID for queries.
         self.tokenizer.pad_token_id = self.tokenizer.mask_token_id
 
-        # Override the configuration values with the provided arguments, if any.
         self.document_length = (
             document_length
             if document_length is not None
@@ -325,6 +393,11 @@ class ColBERT(SentenceTransformer):
         self.skiplist = [
             self.tokenizer.convert_tokens_to_ids(word) for word in self.skiplist_words
         ]
+        self.attend_to_expansion_tokens = (
+            attend_to_expansion_tokens
+            if attend_to_expansion_tokens is not None
+            else self.attend_to_expansion_tokens or False
+        )
 
     @staticmethod
     def load(input_path) -> "ColBERT":
@@ -720,6 +793,35 @@ class ColBERT(SentenceTransformer):
 
         return pooled_embeddings
 
+    @property
+    def similarity_fn_name(self) -> Literal["MaxSim"]:
+        """Return the name of the similarity function used by :meth:`SentenceTransformer.similarity` and :meth:`SentenceTransformer.similarity_pairwise`.
+
+        Returns:
+            Optional[str]: The name of the similarity function. Can be None if not set, in which case it will
+                default to "cosine" when first called.
+        Examples
+        --------
+        >>> model = ColBERT("bert-base-uncased")
+        >>> model.similarity_fn_name
+            'MaxSim'
+        """
+        if self._similarity_fn_name is None:
+            self.similarity_fn_name = SimilarityFunction.MAXSIM
+        return self._similarity_fn_name
+
+    @similarity_fn_name.setter
+    def similarity_fn_name(self, value: Literal["MaxSim"] | SimilarityFunction) -> None:
+        if isinstance(value, SimilarityFunction):
+            value = value.value
+        self._similarity_fn_name = value
+
+        if value is not None:
+            self._similarity = SimilarityFunction.to_similarity_fn(value)
+            self._similarity_pairwise = SimilarityFunction.to_similarity_pairwise_fn(
+                value
+            )
+
     @staticmethod
     def skiplist_mask(input_ids: torch.Tensor, skiplist: list[int]) -> torch.Tensor:
         """Create a mask for the set of input_ids that are in the skiplist."""
@@ -985,46 +1087,6 @@ class ColBERT(SentenceTransformer):
             config["skiplist_words"] = self.skiplist_words
             json.dump(config, fOut, indent=2)
 
-    def _create_model_card(
-        self,
-        path: str,
-        model_name: str | None = None,
-        train_datasets: list[str] | None = None,
-    ) -> None:
-        """
-        Create an automatic model and stores it in the specified path. If no training was done and the loaded model
-        was a Sentence Transformer model already, then its model card is reused.
-
-        Parameters
-        ----------
-        path
-            The path where the model card will be stored.
-        model_name
-            The name of the model.
-        """
-        if model_name:
-            model_path = Path(model_name)
-            if not model_path.exists() and not self.model_card_data.model_id:
-                self.model_card_data.model_id = model_name
-
-        # If we loaded a Sentence Transformer model from the Hub, and no training was done, then
-        # we don't generate a new model card, but reuse the old one instead.
-        if self._model_card_text and self.model_card_data.trainer is None:
-            model_card = self._model_card_text
-        else:
-            try:
-                model_card = generate_model_card(model=self)
-            except Exception:
-                logger.error(
-                    f"Error while generating model card:\n{traceback.format_exc()}"
-                    "Consider opening an issue on https://github.com/UKPLab/sentence-transformers/issues with this traceback.\n"
-                    "Skipping model card creation."
-                )
-                return
-
-        with open(os.path.join(path, "README.md"), "w", encoding="utf8") as fOut:
-            fOut.write(model_card)
-
     def _load_auto_model(
         self,
         model_name_or_path: str,
@@ -1063,7 +1125,7 @@ class ColBERT(SentenceTransformer):
 
         """
         logger.warning(
-            f"No sentence-transformers model found with name {model_name_or_path}. Creating a ColBERT model from base encoder."
+            f"No sentence-transformers model found with name {model_name_or_path}."
         )
 
         shared_kwargs = {
@@ -1116,7 +1178,7 @@ class ColBERT(SentenceTransformer):
         config_kwargs: dict | None = None,
     ) -> list[nn.Module]:
         """Create a Sentence Transformer model from a model name or path."""
-        modules = super()._load_sbert_model(
+        modules, module_kwargs = super()._load_sbert_model(
             model_name_or_path=model_name_or_path,
             token=token,
             cache_folder=cache_folder,
@@ -1162,4 +1224,4 @@ class ColBERT(SentenceTransformer):
             for module in modules.values()
             if isinstance(module, Transformer)
             or isinstance(module, DenseSentenceTransformer)
-        ]
+        ], module_kwargs
